# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2020 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import typing

from sqlalchemy.sql import select, and_, or_

from secrets import token_urlsafe
from datetime import datetime
from uuid import uuid4

from ..resources import Sessions
from ..tables import community, scoreboard_total

from .exceptions import CommunityTaken, AlreadyCommunity, InvalidCommunity, \
    NoOwnership, InvalidAPIKey
from .models import CommunityModel, MatchModel
from .match import Match


class Community:
    def __init__(self, community_name: str) -> str:
        """
        Handles community interactions.

        Paramters
        ---------
        community_name: str
            ID of community.
        """

        self.community_name = community_name

    async def create_match(self, team_1_name: str, team_2_name: str,
                           team_1_side: int, team_2_side: int,
                           team_1_score: int, team_2_score: int,
                           map_name: str) -> Match:
        """
        Creates a match.

        Returns
        -------
        Match
            Used for interacting with matches.
        """

        match_id = str(uuid4())

        query = scoreboard_total.insert().values(
            match_id=match_id,
            team_1_name=team_1_name,
            team_2_name=team_2_name,
            team_1_side=team_1_side,
            team_2_side=team_2_side,
            map=map_name,
            name=self.community_name,
            team_1_score=team_1_score,
            team_2_score=team_2_score,
            status=1,
            demo_status=0,
            timestamp=datetime.now()
        )

        await Sessions.database.execute(query=query)

        return self.match(match_id)

    def match(self, match_id) -> Match:
        """
        Handles interactions with a match

        Paramters
        ---------
        match_id: str
            ID of match
        """

        return Match(match_id, self.community_name)

    async def regenerate(self) -> None:
        """
        Regenerates API key.
        """

        query = community.update().values(
            api_key=token_urlsafe(24)
        ).where(community.c.name == self.community_name)

        await Sessions.database.execute(query=query)

    async def exists(self) -> bool:
        """
        Checks if community exists with name.
        """

        query = community.count().where(
            community.c.name == self.community_name
        )

        return await Sessions.database.fetch_val(query=query) > 0

    async def matches(self, search: str = None,
                      page: int = 1, limit: int = 5
                      ) -> typing.AsyncGenerator[typing.Any, None]:
        """
        Lists matches.

        Paramters
        ---------
        search: str
            Defaults to None, parameters to search.
        page: int
            What page we on booiiiiii.
        limit: int
            Max amount of matches to display at once.

        Yields
        ------
        MatchModel
            Holds basic match details.
        Match
            Used for interacting with a match.
        """

        query = select([
            scoreboard_total.c.match_id,
            scoreboard_total.c.timestamp,
            scoreboard_total.c.status,
            scoreboard_total.c.demo_status,
            scoreboard_total.c.map,
            scoreboard_total.c.team_1_name,
            scoreboard_total.c.team_2_name,
            scoreboard_total.c.team_1_score,
            scoreboard_total.c.team_2_score,
            scoreboard_total.c.team_1_side,
            scoreboard_total.c.team_2_side
        ]).select_from(
            scoreboard_total
        )

        if search:
            like_search = "%{}%".format(search)

            query = query.where(
                and_(
                    scoreboard_total.c.name == self.community_name,
                    or_(
                        scoreboard_total.c.map.like(like_search),
                        scoreboard_total.c.team_1_name.like(like_search),
                        scoreboard_total.c.team_2_name.like(like_search)
                    )
                )
            )
        else:
            query = query.where(
                scoreboard_total.c.name == self.community_name,
            )

        query = query.order_by(
            scoreboard_total.c.timestamp.desc()
        ).limit(limit).offset((page - 1) * limit if page > 1 else 0)

        async for row in Sessions.database.iterate(query=query):
            yield MatchModel(row), self.match(row["match_id"])

    async def get(self) -> CommunityModel:
        """
        Gets base community details.

        Returns
        -------
        CommunityModel
            Holds community data.

        Raises
        ------
        InvalidCommunity
            Raised when community ID doesn't exist.
        """

        query = select([
            community.c.api_key,
            community.c.owner_id,
            community.c.disabled
        ]).select_from(
            community
        ).where(
            community.c.name == self.community_name
        )

        row = await Sessions.database.fetch_one(query=query)

        if row:
            return CommunityModel(row)
        else:
            raise InvalidCommunity()

    async def disable(self) -> None:
        query = community.update().where(
            community.c.name == self.community_name
        ).values(disabled=True)

        await Sessions.database.execute(query=query)


async def api_key_to_community(api_key: str) -> Community:
    """
    Converts API key to community name.

    Raises
    ------
    InvalidAPIKey

    Returns
    -------
    str
        Community name
    """

    query = select([community.c.name]).select_from(
        community
    ).where(
        and_(
            community.c.api_key == api_key,
            community.c.disabled == 0
        )
    )

    row = await Sessions.database.fetch_val(query=query)

    if row:
        return Community(row)
    else:
        raise InvalidAPIKey()


async def get_community_from_owner(steam_id: str) -> Community:
    """
    Gets community name from owners steamID.

    Raises
    ------
    NoOwnership
        Raised when steam id doesn't own any communties.
    """

    query = select(
        [community.c.name]
    ).select_from(
        community
    ).where(
        and_(
            community.c.owner_id == steam_id,
            community.c.disabled == 0
        )
    )

    name = await Sessions.database.fetch_val(query=query)

    if name:
        return Community(name)
    else:
        raise NoOwnership()


async def owner_exists(steam_id: str) -> bool:
    """
    Checks if given steam_id owns a community
    """

    query = community.count().where(
        and_(
            community.c.owner_id == steam_id,
            community.c.disabled == 0
        )
    )

    return await Sessions.database.fetch_val(query=query) > 0


async def create_community(steam_id: str, community_name: str,
                           disabled: bool = False) -> (Community, str):
    """
    Paramters
    ---------
    owner_id: str
        Owner ID.
    name: str
        Name of community.
    disabled: bool
        Defaults to False.

    Returns
    -------
    Community
        Used for interacting with a community
    str
        Edit name of community

    Raises
    ------
    CommunityTaken
        Raised when community name is taken.
    AlreadyCommunity
        Raised when owner already owns a community.
    """

    if await Community(community_name).exists():
        raise CommunityTaken()

    if await owner_exists(steam_id):
        raise AlreadyCommunity()

    query = community.insert().values(
        name=community_name,
        owner_id=steam_id,
        disabled=disabled,
        api_key=token_urlsafe(24)
    )

    await Sessions.database.execute(query=query)

    return Community(community_name)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2020 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette_wtf import CSRFProtectMiddleware

from secrets import token_urlsafe

from databases import Database
from aiohttp import ClientSession

import backblaze

from .tables import create_tables
from .resources import Sessions, Config
from .settings import DatabaseSettings, B2Settings
from .routes import ROUTES, ERROR_HANDLERS
from .http_middleware import APIMiddleware


__version__ = "0.0.9"


MAP_IMAGES = {
    "de_austria": "austria.jpg",
    "de_cache": "cache.jpg",
    "de_canals": "canals.jpg",
    "de_cbble": "cbble.jpg",
    "de_dust": "dust.jpg",
    "de_dust2": "dust2.jpg",
    "de_inferno": "inferno.jpg",
    "de_mirage": "mirage.jpg",
    "de_nuke": "nuke.jpg",
    "de_overpass": "overpass.jpg",
    "de_train": "train.jpg",
}


class SQLMatches(Starlette):
    def __init__(self, database_settings: DatabaseSettings,
                 b2_settings: B2Settings,
                 friendly_url: str,
                 secret_key: str = token_urlsafe(),
                 csrf_secret: str = token_urlsafe(),
                 map_images: dict = MAP_IMAGES,
                 upload_delay: float = 0.1,
                 max_upload_size: int = 80000000,
                 **kwargs) -> None:
        """
        SQLMatches server.

        database_settings: DatabaseSettings
            Holds settings for database.
        friendly_url: str
            URL to project.
        csrf_secret: str
            Optionally pass your own url safe secret key.
        secret_key: str
            Optionally pass your own url safe secret key.
        map_images: dict
            Key as actual map name, value as image name.
        upload_delay: float
            by default 0.1
        max_upload_size: int
            by default 80000000
        kwargs
            Additional parameters to pass to the Starlette instance.
        """

        startup_tasks = [self._startup]
        shutdown_tasks = [self._shutdown]

        if "on_startup" in kwargs:
            startup_tasks = startup_tasks + kwargs["on_startup"]

        if "on_shutdown" in kwargs:
            shutdown_tasks = shutdown_tasks + kwargs["on_shutdown"]

        middlewares = [
            Middleware(SessionMiddleware, secret_key=secret_key),
            Middleware(CSRFProtectMiddleware, csrf_secret=csrf_secret),
            Middleware(APIMiddleware)
        ]
        if "middleware" in kwargs:
            middlewares = middlewares + kwargs["middleware"]

        if "routes" in kwargs:
            routes = kwargs["routes"] + ROUTES
        else:
            routes = ROUTES

        if "exception_handlers" in kwargs:
            exception_handlers = kwargs["exception_handlers"] + ERROR_HANDLERS
        else:
            exception_handlers = ERROR_HANDLERS

        if friendly_url[:1] != "/":
            friendly_url += "/"

        Config.url = friendly_url
        Config.map_images = map_images
        Config.demo_pathway = b2_settings.pathway
        Config.cdn_url = b2_settings.cdn_url
        Config.upload_delay = upload_delay
        Config.max_upload_size = max_upload_size

        database_url = "://{}:{}@{}:{}/{}?charset=utf8mb4".format(
            database_settings.username,
            database_settings.password,
            database_settings.server,
            database_settings.port,
            database_settings.database
        )

        Sessions.database = Database(
            database_settings.engine + database_url
        )

        self.b2 = backblaze.Awaiting(
            b2_settings.key_id,
            b2_settings.application_key
        )

        Sessions.bucket = self.b2.bucket(
            b2_settings.bucket_id
        )

        create_tables(
            "{}+{}{}".format(
                database_settings.engine,
                database_settings.alchemy_engine,
                database_url
            )
        )

        Starlette.__init__(
            self,
            routes=routes,
            exception_handlers=exception_handlers,
            middleware=middlewares,
            on_startup=startup_tasks,
            on_shutdown=shutdown_tasks,
            **kwargs
        )

    async def _startup(self) -> None:
        """
        Starts up needed sessions.
        """

        await Sessions.database.connect()
        await self.b2.authorize()
        Sessions.aiohttp = ClientSession()

    async def _shutdown(self) -> None:
        """
        Closes any underlying sessions.
        """

        await Sessions.database.disconnect()
        await self.b2.close()
        await Sessions.aiohttp.close()



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2020 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from starlette.endpoints import HTTPEndpoint
from starlette.responses import RedirectResponse
from ..templating import TEMPLATE

from urllib.parse import unquote_plus

from ..community import Community, get_community_from_owner
from ..community.exceptions import InvalidCommunity, NoOwnership

from ..resources import Config


class CommunityPage(HTTPEndpoint):
    async def get(self, request):
        community = Community(request.path_params["community"])

        try:
            base_details = await community.get()
        except InvalidCommunity:
            return RedirectResponse("/")
        else:
            matches = [data async for data, _ in community.matches(
                page=request.path_params["page"] if "page" in
                request.path_params else 1,
                search=unquote_plus(request.query_params["search"])
                if "search" in request.query_params else None
            )]

            return TEMPLATE.TemplateResponse(
                "community.html",
                {
                    "request": request,
                    "base_details": base_details,
                    "matches": matches,
                    "map_images": Config.map_images
                }
            )


class CommunityUpdateKeyPage(HTTPEndpoint):
    async def post(self, request):
        if "steam_id" not in request.session:
            return RedirectResponse("/", status_code=303)

        try:
            community = await get_community_from_owner(
                request.session["steam_id"]
            )
        except NoOwnership:
            return RedirectResponse("/", status_code=303)
        else:
            await community.regenerate()

            return RedirectResponse(
                request.url_for(
                    "CommunityPage",
                    community=community.community_name
                ),
                status_code=303
            )


class CommunityDisablePage(HTTPEndpoint):
    async def post(self, request):
        if "steam_id" not in request.session:
            return RedirectResponse("/", status_code=303)

        try:
            community = await get_community_from_owner(
                request.session["steam_id"]
            )
        except NoOwnership:
            return RedirectResponse("/", status_code=303)
        else:
            await community.disable()

            return RedirectResponse(
                request.url_for(
                    "CommunityPage",
                    community=community.community_name
                ),
                status_code=303
            )



import os
import re

from setuptools import setup


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_version(package):
    with open(os.path.join(package, "__init__.py")) as f:
        return re.search(
            "__version__ = ['\"]([^'\"]+)['\"]", f.read()
        ).group(1)


setup(
    name="SQLMatches",
    version=get_version("SQLMatches"),
    url="https://github.com/WardPearce/SQLMatches",
    description="SQLMatches is a free & open source software built around" +
        "giving players & communities easy access to match records & demos.",
    author="WardPearce",
    author_email="wardpearce@protonmail.com",
    install_requires=get_requirements(),
    license="GPL v3",
    packages=[
        "SQLMatches",
        "SQLMatches.api",
        "SQLMatches.community",
        "SQLMatches.forms",
        "SQLMatches.routes"
    ],
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=False
)



# -*- coding: utf-8 -*-

"""
GNU General Public License v3.0 (GPL v3)
Copyright (c) 2020-2020 WardPearce
Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from .exceptions import UnSupportedEngine
from .resources import Config


class DatabaseSettings:
    def __init__(
                self,
                username: str,
                password: str,
                server: str,
                port: int,
                database: str,
                engine: str = "mysql"
                ) -> None:
        """
        A class for creating a URL for the Database.

        Attributes
        ----------
        username: str
            Database username.
        password: str
            Password for user.
        server: str
            Address for database server.
        port: int
            Database server port.
        database: str
            Name of database.
        engine: str
            Database's engine (Defaults 'mysql').
        """

        self.username = username
        self.password = password
        self.server = server
        self.port = port
        self.database = database
        self.engine = engine

        if engine == "mysql":
            self.alchemy_engine = "pymysql"
        elif engine == "sqlite":
            self.alchemy_engine = "sqlite3"
        elif engine == "postgresql":
            self.alchemy_engine = "psycopg2"
        else:
            raise UnSupportedEngine()

        Config.db_engine = engine


class B2Settings:
    def __init__(self, key_id: str, application_key: str,
                 bucket_id: str, pathway: str, cdn_url: str) -> None:
        """
        Stores b2 settings.

        Attributes
        ----------
        key_id: str
            B2 key ID.
        application_key: str
            B2 app key.
        bucket_id: str
            Bucket to upload demos to.
        pathway: str
            Pathway to store demos to.
        cdn_url: str
            URL to access files.
        """

        self.key_id = key_id
        self.application_key = application_key
        self.bucket_id = bucket_id
        self.cdn_url = cdn_url if cdn_url[-1:] == "/" else cdn_url + "/"
        self.pathway = pathway if pathway[-1:] == "/" else pathway + "/"
